/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: image.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: image.c,v 1.4 2002/05/08 21:34:35 jdejong Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Routines for writing standard FITS images                              */
/* I/O is a plane at a time, images are as floats                         */
/*----------------------------------------------------------------------- */
#include "image.h"
#include "math.h"

image_info* 
open_image(char *file, int *ierr)
/*----------------------------------------------------------------------- */
/*  Open an existing FITS image file                                      */
/*   Inputs:                                                              */
/*      file       name of the file to be read                            */
/*   Returns:                                                             */
/*      An image_info pointer needed for I/O; NULL indicates failure      */
/*----------------------------------------------------------------------- */
{
  image_info *out=NULL;
  int i, bp, naxis;  
  long  naxes[3];

  if (file==NULL) return out; /* sanity check */
  if (*ierr>0) return NULL;   /* existing cfitsio error? */

  /* create output object */
  out = (image_info*)malloc(sizeof(image_info));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating image_info for %s\n", file);
    return out;
  }
  memset (out, 0, sizeof(image_info)); /* zero fill */

  /* save file name */
  out->filename = (char*)malloc(strlen(file)+1);
  strcpy (out->filename, file);

  /* attempt to open file */
  fits_open_file(&out->fptr, file, READONLY, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening input FITS cube %s \n", *ierr, 
	    file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    kill_image_info(out); /* cleanup */
    return NULL;
  }/* end error test */

  /* get image info */
  fits_read_imghdr (out->fptr, 3, NULL, &bp, &naxis, naxes, 
		    NULL, NULL, NULL, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d reading FITS cube info %s \n", 
	    *ierr, file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    kill_image_info(out); /* cleanup */
    return NULL;
  }/* end error test */

  /* set image info*/
  out->nplane = 0; /* nothing read yet */

  /* set datatype */
  bp = 16;
  fits_read_key (out->fptr, TINT, "BITPIX", (void*)&bp, NULL, ierr);
  if (bp==8) out->dtype = TBYTE;
	else if (bp==16)
   {
		 float bzero, bscale ;
		 fits_read_key (out->fptr, TFLOAT, "BZERO", (void*)&bzero, NULL, ierr);
		 if( *ierr)
			 bzero=0. ;
		 fits_read_key (out->fptr, TFLOAT, "BSCALE", (void*)&bscale, NULL, ierr);
		 if( *ierr)
			 bscale=1. ;
		 /* a hack: MIDI has bp=16 and BZERO=32756 ! */
		 if( fabs(bzero-32768.)>0.01 || fabs(bscale-1.)>0.001 )
			 out->dtype = TSHORT;
		 else
			 out->dtype = TUSHORT;
   }
  else if (bp==32) out->dtype = TLONG;
  else if (bp==-32) out->dtype = TFLOAT;
  else if (bp==-64) out->dtype = TDOUBLE;
  else out->dtype = TFLOAT;

  out->dim[0] = naxes[0];
  out->dim[1] = naxes[1];
  out->dim[2] = naxes[2];

  return out;  /* apparently OK */
} /* end open_image */

image_info* 
make_image(char *file, int bp, int  naxis[3], char *ctype[3], 
	   double crval[3], float crpix[3], float crinc[3],  
	   float crota[3], char *date_obs, int *ierr)
/*----------------------------------------------------------------------- */
/*   Create/init a FITS image file                                        */
/*   Inputs:                                                              */
/*      file       name of the file to be created                         */
/*      bp         desired bits per pixel in the image                    */
/*      naxis      dimension of two axes of data, weight arrays           */
/*      ctype      labels for the axes, NULL->ignore                      */
/*      crval      reference value for each axis , NULL->ignore           */
/*      crpix      reference pixels, NULL->ignore                         */
/*      crinc      coordinate increments, NULL->ignore                    */
/*      crota      coordinate rotations, NULL->ignore                     */
/*      date_obs   Date (yyyy-mm-dd) observations began, NULL->ignore     */
/*   Returns:                                                             */
/*      An image_info pointer needed for I/O; NULL indicates failure      */
/*----------------------------------------------------------------------- */
{
  int i;  
  image_info *out=NULL;
  char keyword[9];

  if (*ierr!=0) return NULL;  /* existing error? */
  if (file==NULL) return out; /* sanity check */

  /* create output object */
  out = (image_info*)malloc(sizeof(image_info));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating image_info for %s\n", file);
    return out;
  }
  memset (out, 0, sizeof(image_info)); /* zero fill */

  /* save file name */
  out->filename = (char*)malloc(strlen(file)+1);
  strcpy (out->filename, file);

  /* attempt to open/create file */
  fits_create_file(&out->fptr, file, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d opening output FITS file %s \n", *ierr, 
	    file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    kill_image_info(out); /* cleanup */
    return NULL;
  }/* end error test */

  /* initialize image */
  out->dim[0] = naxis[0];
  out->dim[1] = naxis[1];
  out->dim[2] = 1;
  fits_create_img (out->fptr, bp, 3, out->dim, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR %d initializinging output FITS file %s \n", 
	    *ierr, file);
    table_print_error_stack();  /* Show any cfitsio error messages */
    kill_image_info(out); /* cleanup */
    return NULL;
  }/* end error test */

  /* No planes written so far */
  out->nplane = 0;

  /* set datatype */
  if (bp==8) out->dtype = TBYTE;
  else if (bp==16) out->dtype = TSHORT;
  else if (bp==32) out->dtype = TLONG;
  else if (bp==-32) out->dtype = TFLOAT;
  else if (bp==-64) out->dtype = TDOUBLE;
  else out->dtype = TFLOAT;

  /* write descriptive info if given*/
  for (i=0; i<3; i++) {
    if (ctype!=NULL){
      sprintf (keyword,"CTYPE%d", i+1);
      fits_write_key(out->fptr, TSTRING, keyword, (void*)(ctype[i]),
		     "Axis label", ierr);
    }
    if (crval!=NULL){
      sprintf (keyword,"CRVAL%d", i+1);
      fits_write_key(out->fptr, TDOUBLE, keyword, (void*)(&crval[i]),
		     "Coordinate value at ref. pixel", ierr);
    }
    if (crpix!=NULL){
      sprintf (keyword,"CRPIX%d", i+1);
      fits_write_key(out->fptr, TFLOAT, keyword, (void*)(&crpix[i]),
		     "Coordinate reference pixel", ierr);
    }
    if (crinc!=NULL){
      sprintf (keyword,"CDELT%d", i+1);
      fits_write_key(out->fptr, TFLOAT, keyword, (void*)(&crinc[i]),
		     "Coordinate increment", ierr);
    }
    if (crota!=NULL){
      sprintf (keyword,"CROTA%d", i+1);
      fits_write_key(out->fptr, TFLOAT, keyword, (void*)(&crota[i]),
		     "Coordinate rotation", ierr);
    }

  } /* end loop over descriptive parameters */
  if (date_obs!=NULL)
    fits_write_key(out->fptr, TSTRING, "DATE-OBS", (void*)(date_obs),
      "Observing date (yyyy-mm-dd)", ierr);

  return out;  /* apparently OK */
} /* end make_image */

void kill_image_info (image_info *info)
/*----------------------------------------------------------------------- */
/*   Delete an image_info structure, should already have been closed.     */
/*   Inputs:                                                              */
/*      row    pointer to the image_info structure.                       */
/*----------------------------------------------------------------------- */
{
  if (info==NULL) return;                         /* anybody home? */
  if (info->filename!=NULL) free(info->filename); /* file name */
  free (info);                                    /* free basic structure */
} /* end kill_image_info */


int image_read(image_info *info, float *plane, int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next plane in an image                                      */
/*   Inputs:                                                              */
/*      info      image_info for I/O                                      */
/*      plane     pointer plane data, will be converted as necessary      */
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  long  first, number;
  int   anynull=0;
  float null=-1.0;

  if (info==NULL)  return 1;  /* anybody home? */
  if (plane==NULL) return 1;  /* anybody home? */
  if (*ierr!=0) return 1;     /* existing error? */

  /* where in file to read */
  number = info->dim[0] * info->dim[1];
  first = 1 + number * info->nplane++;

  return fits_read_img (info->fptr, TFLOAT, first, number, 
		  &null, (void*)plane, &anynull, ierr);
} /* end image_read */

int image_read_raw(image_info *info, void *plane, int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next plane in an image with no type conversion              */
/*   No checks are make for invalid data.                                 */
/*   Inputs:                                                              */
/*      info      image_info for I/O                                      */
/*      plane     pointer plane data, in file format                      */
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  long  first, number;
  int   anynull=0;

  if (info==NULL)  return 1;  /* anybody home? */
  if (plane==NULL) return 1;  /* anybody home? */
  if (*ierr!=0) return 1;     /* existing error? */

  /* where in file to read */
  number = info->dim[0] * info->dim[1];
  first = 1 + number * info->nplane++;

  return fits_read_img (info->fptr, info->dtype, first, number, 
		  NULL, (void*)plane, &anynull, ierr);
} /* end image_read_raw */

int image_write(image_info *info, float *plane, int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next plane into an image                                  */
/*   Inputs:                                                              */
/*      info      image_info for I/O                                      */
/*      plane     pointer plane data, will be converted as necessary      */
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  long first, number;

  if (info==NULL)  return 1;  /* anybody home? */
  if (plane==NULL) return 1;  /* anybody home? */
  if (*ierr!=0) return 1;      /* existing error? */

  /* where in file to write */
  number = info->dim[0] * info->dim[1];
  first = 1 + number * info->nplane++;

  return fits_write_img (info->fptr, TFLOAT, first, number, 
		  (void*)plane, ierr);
} /* end image_write */

int image_close(image_info *info, int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an image updating the header                                   */
/*   Inputs:                                                              */
/*      info      image_info for I/O                                      */
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (info==NULL)  return 1;  /* anybody home? */
  if (*ierr!=0) return 1;      /* existing error? */

  /* update header with number of planes */
  fits_update_key (info->fptr, TLONG, "NAXIS3", (void*)(&info->nplane), 
		   "Number of planes in image", ierr);

  /* close it */
  return fits_close_file (info->fptr, ierr);
} /* end image_close */
/* */
